#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 23 19:03:10 2020

@author: jiayingweng
"""

import numpy as np
from scipy import stats
import scipy.linalg as la
import FTIRE

__all__ = ['SIR']

class SIR:
    """SIR method to apply for X and Y"""
    def __init__(self, X, y, d = 1, n_slice = 10, spX = False):
        self.d = d
        self.n_slice = n_slice
        self.X = np.array(X)
        self.y = np.array(y)
        self.M = None
        self.B = None
        self.values = None
        self.pval = None
        self.test = None
        self.spX = spX
    
    def __str__(self):
        return "This is SIR method applied for {} dimension reduction with {} slices.".format(self.d, self.n_slice)
    
    def dr(self, d = None):
        if d is None:
            d = self.d
        n, p = self.X.shape
        ## kernel matrix
        self.kernel()
        ## covariance matrix
        # n = self.X.shape[0]
        covX = FTIRE.SpCov.spcovCV(self.X) if self.spX else np.cov(self.X.T)
        ## SVD
        if p == 1:
            self.B = np.ones((p,d))
        elif n < p:
            self.B = np.zeros((p,d))
        else:
            self.values, vectors = la.eigh(self.M, covX)
            decsort = np.argsort(self.values)[::-1]
            self.B = vectors[:,decsort[range(d)]]
        return covX
    
    def upd(self):
        """ update dimension d """
        self.testd()
        self.d = next((i for i, j in enumerate(self.pval > 0.05) if j), None)
        if self.d is None:
            self.d = self.X.shape[1]
        self.dr(self.d)
        return self.B 
    
    def transformX(self):
        self.dr(self.d)
        newX = self.X @ self.B
        return newX
    
    def kernel(self):
        n, p = self.X.shape
        X0 = self.X - np.mean(self.X, axis = 0)
        # slice y into n_slice
        YI = np.argsort(self.y.reshape(-1))
        split = np.array_split(YI, self.n_slice)
        
        ## kernel matrix
        self.M = np.zeros((p, p))
        Ups = np.zeros((p, self.n_slice))
        for i in range(self.n_slice):
            if p == 1:
                tt = X0[split[i]].reshape(-1, p)
            else:
                tt = X0[split[i],].reshape(-1, p)   
            Xh = np.mean(tt, axis = 0).reshape(p, 1)
            Ups[:,i] = np.sqrt(len(split[i])/n) * Xh.reshape(-1)
            self.M = self.M + Xh @ Xh.T * len(split[i])/n
        return Ups
    
    
    def testd(self):
        n, p = self.X.shape
        ## kernel matrix
        self.kernel()
        ## covariance matrix
        covX = FTIRE.SpCov.spcovCV(self.X) if self.spX else np.cov(self.X.T)
        ## SVD
        self.values, vectors = la.eigh(self.M, covX)
        ## Sequential Tests
        decsort = np.argsort(np.abs(self.values))[::-1]
        self.pval = self.test = np.zeros(p)
        for r in range(p):
            self.test[r] = n * np.sum(self.values[decsort[r:p]])
            self.pval[r] = 1 - stats.chi2.cdf(self.test[r], (p-r)*(self.n_slice-1-r))
        return covX

    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 23 21:53:32 2020

@author: jiayingweng
"""

import numpy as np
from scipy import stats
import scipy.linalg as la
import FTIRE

__all__ = ['FT']
#%%

class FT:
    """FT method to apply for X and Y"""
    def __init__(self, X, y, d = 1, m = 30, weighted = False, W = None, spX = False):
        self.d = d
        self.m = m
        self.X = np.array(X)
        self.y = np.array(y)
        self.M = None
        self.B = None
        self.values = None
        self.pval = None
        self.test = None
        self.weighted = weighted
        self.W = W
        self.spX = spX

    
    def __str__(self):
        return "This is FT method (standardized f(y)) applied for {} dimension reduction with m={} and weight={}.".format(self.d, self.m, self.weighted)
    
    def dr(self, d = None):
        if d is None:
            d = self.d
        n, p = self.X.shape
        ## kernel matrix
        self.kernel()
        ## covariance matrix
        # n = self.X.shape[0]
        covX = FTIRE.SpCov.spcovCV(self.X) if self.spX else np.cov(self.X.T)
        ## SVD
        if p == 1:
            self.B = np.ones((p,d))
        elif n < p:
            self.B = np.zeros((p,d))
        else:
            self.values, vectors = la.eigh(self.M, covX)
            decsort = np.argsort(self.values)[::-1]
            self.B = vectors[:,decsort[range(d)]]
        return covX
    
    def upd(self):
        """ update dimension d """
        self.testd()
        self.d = next((i for i, j in enumerate(self.pval > 0.05) if j), None)
        if self.d is None:
            self.d = self.X.shape[1]
        self.dr(self.d)
        return self.B
    
    def transformX(self):
        self.dr(self.d)
        newX = self.X @ self.B
        return newX
    
    
    def kernel(self, weighted=None):
        if weighted is None:
            weighted = self.weighted
        n, p = self.X.shape
        self.y = self.y.reshape(n,-1)
        q = self.y.shape[1]
        X0 = self.X - np.mean(self.X, axis = 0)
        if self.W is None:
            sig = 0.1 * np.pi ** 2/ np.median(la.norm(self.y, axis = 1) **2)
            self.W = np.random.rand(q, self.m) * sig
        # Mxy = np.zeros((p, 2*self.m))
        My = np.zeros((n, 2*self.m))
        for i in range(self.m):
            w = self.W[:,i]
            if q > 1:
                w = w/( np.linalg.norm(self.W[:,i]) * np.ones((1,q)) )
            yw =  self.y @ w.T #* 2 * np.pi      
            cosy = np.array(np.cos(yw))
            cosy = (cosy - np.mean(cosy, axis = 0))/np.std(cosy, axis = 0)
            My[:, 2*i] = np.array(cosy).reshape(-1)
            siny = np.sin(yw)
            siny = (siny - np.mean(siny, axis = 0))/np.std(siny, axis = 0)
            My[:, 2*i + 1] = np.array(siny).reshape(-1)
        
        sigcov = np.cov(np.concatenate((X0, My), axis = 1),rowvar=False)
        sigxf = sigcov[0:p, p:]
        sigff = sigcov[p:, p:]
        
        if weighted:
            ivcovy = self.matpower(cov = sigff, a = -1)
            self.M = sigxf @ ivcovy @ sigxf.T
        else:
            self.M = sigxf @ sigxf.T
        
        return sigxf
    
    def testd(self):
        n, p = self.X.shape
        ## kernel matrix
        self.kernel()
        ## covariance matrix
        covX = FTIRE.SpCov.spcovCV(self.X) if self.spX else np.cov(self.X.T)
        ## SVD
        self.values, vectors = la.eigh(self.M, covX)
        ## Sequential Tests
        decsort = np.argsort(np.abs(self.values))[::-1]
        self.pval = self.test = np.zeros(p)
        for r in range(p):
            self.test[r] = n * np.sum(self.values[decsort[r:p]])
            self.pval[r] = 1 - stats.chi2.cdf(self.test[r], (p-r)*(2*self.m-r))
        return covX

    
def matpower(cov, a, err=1e-3):
    value, vector = la.eigh(cov)
    indv = value > err
    trcov = vector[:,indv] @ np.diag(np.power(value[indv], a)) @ vector[:,indv].T
    return trcov
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jan 23 15:59:09 2021

@author: jiayingweng
"""
__all__ = ["ftire","FT","SIR","genxy","SpCov"]


from . import genxy
from . import SpCov
from FTIRE.SpCov import spcovCV
from . import FT
from . import SIR
from FTIRE.SIR import SIR as sir
from FTIRE.FT import FT as ft
from . import ftire
from FTIRE.ftire import *




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 28 16:21:46 2021

@author: jiayingweng
"""

import numpy as np
import scipy.linalg as la

__all__ = ['generateX', 'generateY']

def generateX(n, p, covstr):  
    """
    Generate X for simulation
    Args:
        n (int): sample size
        p (int): number of dimension of X
        covstr (0-3): covariance structure
    Returns:
        X: n times p array
    """
    ## generate X 
    if covstr == 0:
        covx = np.eye(p)
    elif covstr == 1:
        v = 0.5 ** np.arange(p)
        covx = la.toeplitz(v)
    elif covstr == 2:
        offdiag = 0.2
        covx = np.ones((p,p)) * offdiag
        covx = covx + np.eye(p) * (1-offdiag)  
    elif covstr == 3:
        v = 0.8 ** np.arange(p)
        covx = la.toeplitz(v)

    L = np.linalg.cholesky(covx)
    Z = np.random.randn(p,n)
    X = (L @ Z).T  
    return(X)

def generateY(X, M):
    """
    Generate Y based on X
    Args: 
        X: input covariate
        M: model 1-7 uni; 10-15 multi
    Returns:
        Y: outcome
        d: structural dimension
        p: the dimension of Y
        b: the true beta
    """
    [n,p] = X.shape
    ## generate Y      
    if M == 1: # Qian M1
        d = 1
        q = 1
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index = np.arange(5)
        b[index,:] = 1
        y[:,0] = np.exp(X @ b[:,0]) + np.random.randn(n)
    elif M == 2: # Qian M2 
        d = 2
        q = 1
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index1 = np.arange(4) #np.random.randint(p, size = 5)
        index2 = np.arange(p-4,p)
        b[index1,0] = 1
        b[index2, 1] = 1
        y[:,0] = np.sign(X @ b[:,0]) * np.log( np.abs( X @ b[:,1] + 5 ) ) + 0.2 * np.random.randn(n)
    elif M == 3: # Tan AOS Model 1
        d = 1
        q = 1
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index = np.arange(5)
        b[index,:] = 1
        y[:,0] = np.sin(X @ b[:,0]) ** 2 + X @ b[:,0] + np.random.randn(n)
    elif M == 4: # Tan AOS Model 2
        d = 1
        q = 1
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index = np.arange(5)
        b[index,:] = 1
        y[:,0] = 2 * np.tanh(X @ b[:,0]) + np.random.randn(n)
    elif M == 5: # Cook Zhang 
        d = 1
        q = 1
        b = np.zeros((p,d))
        index = np.arange(1)
        b[index,:] = 1
        X = 1/4 * np.sqrt(0.1) * ( np.random.randn(p,n) + 1) + 1/2 * np.sqrt(0.1) * ( np.random.randn(p,n) + 2 ) + 1/4 * np.sqrt(10) * (np.random.randn(p,n) + 1) 
        X = X.T
        y = np.abs( np.sin( X @ b[:,0] ) ) + 0.2 * np.random.randn(n)
    elif M == 6:
        d = 2
        q = 1
        b = np.zeros((p,d))
        b[0,0] = 1
        b[1,1] = 1
        X[:,1] = X[:,0] + X[:,1]
        X[:,3] = ( 1+X[:,1] ) * X[:,3]
        y = X @ b[:,0] + 0.5 * (X @ b[:,1])** 2   
    elif M == 7:
        d = 2
        q = 1
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index1 = np.arange(1)
        index2 = np.arange(1,3)
        b[index1,0] = 1
        b[index2, 1] = 1
        y = (X @ b[:,0]) * (X @ b[:,1] + 1) + np.random.randn(n)
    elif M == 10:
        ## simple
        d = 2
        q = 3
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        #index = np.random.randint(p, size = 5)
        index = np.arange(5)
        b[index[0:2], 0] = 1
        b[index[2:], 1] = 1
        y[:,0] = np.exp( X @ b[:,0]) + 0.5 * np.random.randn(n)
        y[:,1] = X @ b[:,1] + 0.1 * np.random.randn(n)
        y[:,2] = 0.1 * np.random.randn(n)
    elif M == 11: ## Zhu Zhu wen 2010 Example 3
        ## complex
        d = 2
        q = 5
        
        covy = np.diag([1,1/2,1/2,1/3,1/4])
        covy[0,1] = covy[1,0] = -1/2
        L = np.linalg.cholesky(covy)
        Z = np.random.randn(q,n)
        eps = (L @ Z).T
    
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index = np.arange(3) #np.random.randint(p, size = 5)
        b[index[0], 0] = 1
        b[index[1:], 1] = 1
        y[:,0] = 1 + X @ b[:,0] + np.sin(X @ b[:,1]) +  eps[:,0]
        y[:,1] = X @ b[:,1] / (0.5 + (X @ b[:,0])**2) + eps[:,1]
        y[:,2] = np.abs(X @ b[:,1]) * eps[:,2]
        y[:,3] = eps[:,3]
        y[:,4] = eps[:,4]
    elif M == 12: ## Zhu Zhu wen 2010 Example 2 and Bing Li 2008 Model 4.3
        d = 1
        q = 2
        b = np.zeros((p,d))
        b[0:2,0] = [0.8, 0.6]
        top = np.ones((n,2))
        top[:,1] = np.sin(X @ b[:,0])
        y = np.zeros((n,q))
        for i in range(n):
            covy = la.toeplitz(top[i,:])
            L = np.linalg.cholesky(covy)
            Z = np.random.randn(q,1)
            y[i,:] = (L @ Z).T
    elif M == 13: # Bing Li, Weng, Li 2008 Model 4.1
        d = 2
        q = 4
        covy = np.diag([1,1,1,1])
        covy[0,1] = covy[1,0] = -1/2
        L = np.linalg.cholesky(covy)
        Z = np.random.randn(q,n)
        eps = (L @ Z).T
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index = range(3)
        b[index[0:1], 0] = 1
        b[index[1:], 1] = [2,1]
        
        y[:,0] = X @ b[:,0] + eps[:,0]
        y[:,1] = X @ b[:,1] + eps[:,1]
        y[:,2] = eps[:,2]
        y[:,3] = eps[:,3]
    elif M == 14: # Bing li 2008 Model 4.2
        d = 1
        q = 4
        b = np.zeros((p,d))
        b[0:2,0] = [0.8, 0.6]
        top = np.sin(X @ b[:,0])
        y = np.zeros((n,q))
        for i in range(n):
            covy = np.eye(q)
            covy[0,1] = covy[1,0] = top[i]
            L = np.linalg.cholesky(covy)
            Z = np.random.randn(q,1)
            eps = (L @ Z).T
            y[i,:] = eps
            y[i,0] = np.exp(eps[0,0])
    elif M == 15: # Bing Li 08 Model 4.4
        d = 2
        q = 5
        
        covy = np.diag([1,1/2,1/2,1/3,1/4])
        covy[0,1] = covy[1,0] = -1/2
        L = np.linalg.cholesky(covy)
        Z = np.random.randn(q,n)
        eps = (L @ Z).T
    
        b = np.zeros((p,d))
        y = np.zeros((n,q))
        index = np.arange(5) #np.random.randint(p, size = 5)
        b[index[0:2], 0] = 1
        b[index[2:], 1] = 1
        y[:,0] = X @ b[:,0] + X @ b[:,1] / (0.5 + (X @ b[:,0])**2) + eps[:,0]
        y[:,1] = X @ b[:,0] + np.exp( 0.5 * X @ b[:,1]) +  eps[:,1]
        y[:,2] = X @ b[:,0] + X @ b[:,1] + eps[:,2]
        y[:,3] = eps[:,3]
        y[:,4] = eps[:,4]
    return y, d, q, b



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 24 01:24:00 2020

@author: jiayingweng
"""

import numpy as np
import scipy.linalg as la
from sklearn.model_selection import train_test_split

__all__ = ['spcov', 'spcovCV']

# covxx = np.cov(X.T)
def spcov(covxx, lam, standard):    
    # covxx = np.cov(X.T) * (n-1)/n
    # lam = 0.1
    # standard = True    
    if standard:
        covdiag = np.diag(covxx)
        dhat = np.diag(np.sqrt(covdiag))
        dhatinv = np.diag(1/np.sqrt(covdiag))
        S = dhatinv @ covxx @ dhatinv
    else: 
        S = covxx      
        
    tmp = np.abs(S) - lam
    tmp = (tmp > 0) * tmp
    Ss = np.sign(S) * tmp
    covss = Ss - np.diag(np.diag(Ss)) + np.diag(np.diag(S))
    
    if standard:
        covss = dhat @ covss @ dhat
        
    return covss

def spcovCV(X, standard = False, lambseq = None, K = 10, ntr = None):
    n = X.shape[0]
    covxx = np.cov(X.T) * (n-1)/n
    if ntr is None:
        ntr = np.floor(n*(1-1/np.log(n)))/n
    if standard and (lambseq is None):
        lamseq = np.linspace(0, 0.95, 20)
    elif lambseq is None:     
        abscov = np.abs(covxx)
        lammax = np.max(abscov - np.diag(np.diag(abscov)))
        lamseq = np.linspace(0, lammax, 20)
 
    cvloss = np.zeros((K, 20))
    for k in range(K):
        # print('Fold-', k)
        X_train, X_test = train_test_split(X, test_size=ntr, shuffle = True)
        n_train = X_train.shape[0]
        n_test = X_test.shape[0]
        ss_train = np.cov(X_train.T) * (n_train-1)/n_train
        ss_test = np.cov(X_test.T) * (n_test-1)/n_test
        for i in range(20):
            outcov = spcov(ss_train, lamseq[i], standard)
            cvloss[k, i] = la.norm(outcov-ss_test) ** 2 
            
    l_mean = np.mean(cvloss, axis = 0)
    lambcv = lamseq[np.argmin(l_mean)]
    covss = spcov(covxx, lambcv, standard)
    return covss



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 28 10:11:34 2021

@author: jiayingweng
"""

import numpy as np
import scipy.linalg as la
import sys
import dcor
from sklearn.model_selection import KFold # import KFold
import FTIRE

ft = FTIRE.FT.FT
sir = FTIRE.SIR.SIR

__all__ = ["estimate", "CV"]


def projloss(B, b):
    """
    Projection Distance between B and b matrices
    Args:
        B: p times d matrix
        b: p times d estimated matrix
    Returns:
        projection norm loss
    """
    if np.linalg.cond(B) < 1/sys.float_info.epsilon and np.linalg.cond(b) < 1/sys.float_info.epsilon:
#np.isfinite(np.linalg.cond(B)) and np.isfinite(np.linalg.cond(b)):
        loss = la.norm(B @ la.solve(B.T @ B, B.T) - b @ la.solve(b.T @ b, b.T), 2)
    else:
        loss = 100
    return(loss)

def corrloss(B, b, sigt = None, sigs = None):
    """
    distance correlation between B and b
    Args: 
        B: p times d matrix
        b: p times d estimated matrix
    Returns:
        distance correlation
    """
    if np.linalg.cond(B) < 1/sys.float_info.epsilon and np.linalg.cond(b) < 1/sys.float_info.epsilon: #np.isfinite(np.linalg.cond(B)) and np.isfinite(np.linalg.cond(b)):
        if sigt is None:
            p = b.shape[0]
            sigt = np.diag(np.ones(p))
        if sigs is None:
            p = B.shape[0]
            sigs = np.diag(np.ones(p))
        ## normalize B and b
        B = B @ la.inv(la.sqrtm(B.T @ sigs @ B))
        b = b @ la.inv(la.sqrtm(b.T @ sigt @ b))     
        d = B.shape[1]
        loss = np.trace(la.solve(B.T @ sigs @ B, B.T @ sigs @ b) @ la.solve(b.T @ sigs @ b, b.T @ sigs @ B))/d
    else: 
        loss = 0
    return(loss)

def updateC(B, Ups):
    ## update C
    UpsB = (Ups).T @ B
    W1, D, W2 = la.svd(UpsB, full_matrices=False)
    C = W2 @ W1.T ##!
    return(C)


def estimate(X, y, d, m, lamb, method = "ft", NoB = 5, NoC = 20, NoW=2, spX=False, standard=False):
    """
    Estimate B
    Args:
        X: covariates
        y: outcome
        d: structural dimension
        m: number of transfroms
        lamb: regularization parameter
        method: "ft" or "sir"
        NoB: number of iterate over B within ADMM
        NoC: number of iterate over C
        NoW: number of updating weights
        spX: sparse X or not
        standard: standardize X or not
    Returns:
        B: estimate
        covxx: covariance matrix of X
        err2: differences between objective functions since last step
    """
    
    ## init B, C, Ups
    [n,p] = X.shape
    sdr = sir(X = X, y = y, d = d, n_slice = m) if method=='sir' else ft(X = X, y = y,d = d, m = m)
    Ups = sdr.kernel() # phi: p X 2m, My: n X 2m
    # initial B
    # ft.kernel()
    M = sdr.M
    _, B = la.eigh(M, eigvals = (p-d,p-1))
    ## update C
    C = updateC(B, Ups)
    
    ## Covariance
    covxx = FTIRE.SpCov.spcovCV(X, standard = standard) if spX else np.cov(X.T)
    ## initial weight
    weight = np.ones(p)
    
    def updateB(C, weight, lamb):
        rho = 1
        Z = np.zeros((p,d))
        U = np.zeros((p,d))
        for i in range(NoB):
            ## update B
            B = la.solve(covxx + rho * np.eye(p), Ups @ C.T + rho*Z - rho*U) #speed up 
            ## update Z
            Zold = Z.copy()
            lambj = lamb * weight 
            K = np.maximum(1 - lambj/(rho* la.norm(B + U, axis = 1)), np.zeros(p))
            Z = np.diag(K) @ (B+U)
            ## update U
            U = U + (B-Z) 
            #stop criteria
            epsabs = 1e-4
            epsrel = 1e-4
            epspri = np.sqrt(p) * epsabs + epsrel * max(la.norm(B), la.norm(Z))
            epsdual = np.sqrt(p) * epsabs + epsrel * la.norm(U)            
            # err = max(la.norm(B-Z), la.norm(rho*(Zold-Z)))
            # if i == 99:
            #     print('admm %i times' %100)
            if (la.norm(B-Z) < epspri) and (la.norm(rho*(Zold-Z)) < epsdual):
                # print('admm %i times' %i)
                break  
        return(Z)
    

    err2 = np.zeros((NoW,NoC+1))
    for j in range(NoW):
        oldB = B
        err2[j, 0] = np.trace( - Ups.T @ B @ C + C.T @ B.T @ covxx @ B @ C) + lamb * np.dot(weight[np.logical_not(np.isinf(weight))], la.norm(B[np.logical_not(np.isinf(weight)),:], axis = 1))
        for k in range(NoC): 
            ## update B
            B = updateB(C, weight, lamb)
            ## update C
            C = updateC(B, Ups)
            ## stop criteria 
            err2[j, k+1] = np.trace( - Ups.T @ B @ C + C.T @ B.T @ covxx @ B @ C) + lamb * np.dot(weight[np.logical_not(np.isinf(weight))], la.norm(B[np.logical_not(np.isinf(weight)),:], axis = 1))
            # err2 = np.append(err2, np.trace( - Ups.T @ B @ C + C.T @ B.T @ covxx @ B @ C) + lamb * np.dot(weight[np.logical_not(np.isinf(weight))], la.norm(B[np.logical_not(np.isinf(weight)),:], axis = 1)) )
            
            
            if np.isclose(err2[j, k], 0):
                break
            else:
                if abs(err2[j, k+1] - err2[j, k])/abs(err2[j, k]) < 1e-4:
                    break
        # update weight
        Bnorm = la.norm(B, axis = 1)
        close0 = np.isclose(Bnorm, np.zeros(p))
        weight[close0] = np.inf
        no0 = np.logical_not(close0)
        weight[no0] = (1/Bnorm[no0]) ** 1/2
        weight[no0] = weight[no0]/min(weight[no0]) if sum(no0) > 1 else weight[no0]
        # stop criteria for updating weight
        if np.isclose(la.norm(oldB), 0):
            break
        else: 
            if projloss(oldB,B)/la.norm(oldB) < 1e-4:
                # print('update weight %i times' %j)
                break
    return B, covxx, err2

# cross validation
def CV(X, y, d, m, method="ft", nolamb = 50, nofold=10, NoB = 5, NoC = 20, NoW=2, spX=False, standard=False):  
    """
    Estimate B using the best lambda with cross-validation
    Args:
        X: covariates
        y: outcome
        d: structural dimension
        m: number of transfroms
        method: "ft" or "sir"
        nolamb: the number of lambda
        nofold: the number of fold
        NoB: number of iterate over B within ADMM
        NoC: number of iterate over C
        NoW: number of updating weights
        spX: sparse X or not
        standard: standardize X or not
    Returns:
        B: estimate
        covxx: covariance matrix of X
        lambcv: best lambda
        maximized loss 
    """
    ## par.
    method = 'sir' # or 'ft'
    nofold = 10
    nolamb = 50
    spX = False 
    standard = False
    NoB = 5
    NoC = 20
    NoW = 2
    
    ## generate lambda candidate
    lambmax = 1 #np.max(sdr0.M)/10
    lambmin = lambmax/1000 if method == 'sir' else lambmax/10
    lambseq = np.exp(np.linspace(np.log(lambmin), np.log(lambmax), num=nolamb))
    
    kf = KFold(n_splits=nofold)
    cvloss = np.zeros((nofold, nolamb))
    k = 0
    for train_index, test_index in kf.split(X):
        print('Fold-', k)
        
        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]
        #print("TRAIN:", (X_train.shape), "TEST:", (X_test.shape))
        
        for i in range(nolamb):
            Btrain = estimate(X_train,
                              y_train,
                              d,
                              m,
                              lambseq[i],
                              method,
                              NoB,
                              NoC,
                              NoW,
                              spX,
                              standard
                              )[0] # estimate(X, y, d, m, lamb, method = "ft", NoB = 5, NoC = 20, NoW=2, spX=False, standard=False)
            if np.linalg.cond(Btrain) < 1/sys.float_info.epsilon:
                sigs = np.cov(X_train.T)
                Btrain = Btrain @ la.inv(la.sqrtm(Btrain.T @ sigs @ Btrain))
                cvloss[k, i] = 1 - dcor.distance_correlation(X_test @ Btrain, y_test)/d
            else: 
                cvloss[k, i] = 100
        k = k + 1
    l_mean = np.mean(cvloss, axis = 0)
    lambcv = lambseq[np.argmin(l_mean)]
    B, covxx, err= estimate(X, y, d, m, lambcv, method, NoB, NoC, NoW, spX, standard)
    return B, covxx, lambcv, np.argmax(l_mean)

